import weakref

class Node(object): 
    def __init__(self, *args):
        self.branches = list(args)
        self.selectable = False
        
    def accept(self, visitor):
        visitor.push_state(self)
        getattr(visitor, 'visit_' + self.__class__.__name__)(self)
        for node in self.branches:
            node.accept(visitor)
        visitor.pop_state()
        
    def add(self, *args):
        for arg in args:
            arg.parent = weakref.proxy(self)
            self.branches.append(arg)
        
    def remove(self, *args):
        for arg in args:
            self.branches.remove(arg)
    
    def update(self, **kw):
        self.__dict__.update(kw)
        

class Leaf(object):
    pass
    

class Root(Node):
    """
    The root node. There must only be one.
    """
    def __init__(self, *args):
        Node.__init__(self, *args)
        self.background_color = (0.0,0.0,0.0,1.0)
    

class Viewport(Node):
    """
    The Viewport specifies an area of the screen which all children are 
    rendered to.
    """
    def __init__(self, *args):
        Node.__init__(self, *args)
        self.screen_dimensions = (0, 0, 800, 600)
        self.aspect = 800.0 / 600.0
        

class Transform(Node):
    """
    A Transform node is used to scale, move and rotate its child nodes.
    """
    def __init__(self, *args):
        Node.__init__(self, *args)
        self.scale = (1.0,1.0,1.0)
        self.angle = 0.0
        self.axis = (0.0,0.0,1.0)
        self.translate = (0.0,0.0,0.0)
    
    
class Group(Node):
    """
    A Group node is the node which contains renderable items.
    """
    def __init__(self, *args):
        Node.__init__(self, *args)
        self.leaves = []
        
    def add(self, leaf):
        self.leaves.append(leaf)
        
    def remove(self, leaf):
        self.leaves.remove(leaf)
        

class Texture(Leaf):
    """
    A Texture leaf is added to a group node to change the texture which is 
    used to render subsequent quads.
    """
    def __init__(self, filename, filter=True, mipmap=True):
        self.filename = filename
        self.filter = filter
        self.mipmap = mipmap


class Color(Leaf):
    """
    The Color leaf sets the color (including alpha value) for a nodes 
    children and leaves.
    """
    def __init__(self, rgba=(1.0,1.0,1.0,1.0)):
        self.rgba = rgba


class Quad(Leaf):
    """
    The Quad leaf draws a rectangle using the current texture and color. It 
    is drawn from its centre point.
    """
    def __init__(self, (w, h)):
        x,y = w*0.5,h*0.5
        self.vertices = (-x,-y),(-x,y),(x,y),(x,-y)
        
    
class Text(Leaf):
    """
    The Text leaf draws text using a specified font from the bottom left 
    corner.
    """
    def __init__(self, text, font, foreground=(1,1,1), background=None, size=32):
        self.text = text
        self.foreground = foreground
        self.background = background
        self.size = size
        self.font = font
